import abc
import asyncio
import logging
import os
import socket
import ssl
from typing import Any, Awaitable, Callable, List, Optional, Tuple, Type, Union

from asgiref.typing import ASGIApplication
from uvicorn.config import (
    SSL_PROTOCOL_VERSION, Config, HTTPProtocolType, InterfaceType, LifespanType,
    WSProtocolType,
)
from uvicorn.server import Server

from aiomisc.service import Service


log = logging.getLogger(__name__)

UvicornApplication = Union[ASGIApplication, Callable]


class UvicornService(Service, abc.ABC):
    __async_required__: Tuple[str, ...] = (
        "start",
        "create_application",
    )

    sock: Optional[socket.socket] = None

    host: str = "127.0.0.1"
    port: int = 8080
    uds: Optional[str] = None
    fd: Optional[int] = None
    http: Union[Type[asyncio.Protocol], HTTPProtocolType] = "auto"
    ws: Union[Type[asyncio.Protocol], WSProtocolType] = "auto"
    ws_max_size: int = 16 * 1024 * 1024
    ws_ping_interval: Optional[float] = 20.0
    ws_ping_timeout: Optional[float] = 20.0
    ws_per_message_deflate: bool = True
    lifespan: LifespanType = "auto"
    access_log: bool = True
    interface: InterfaceType = "auto"
    proxy_headers: bool = True
    server_header: bool = True
    date_header: bool = True
    forwarded_allow_ips: Optional[Union[List[str], str]] = None
    root_path: str = ""
    limit_concurrency: Optional[int] = None
    limit_max_requests: Optional[int] = None
    backlog: int = 2048
    timeout_keep_alive: int = 5
    timeout_notify: int = 30
    timeout_graceful_shutdown: Optional[int] = None
    callback_notify: Optional[Callable[..., Awaitable[None]]] = None
    ssl_keyfile: Optional[str] = None
    ssl_certfile: Optional[Union[str, os.PathLike]] = None
    ssl_keyfile_password: Optional[str] = None
    ssl_version: int = SSL_PROTOCOL_VERSION
    ssl_cert_reqs: int = ssl.CERT_NONE
    ssl_ca_certs: Optional[str] = None
    ssl_ciphers: str = "TLSv1"
    headers: Optional[List[Tuple[str, str]]] = None
    factory: bool = False
    h11_max_incomplete_event_size: Optional[int] = None

    @abc.abstractmethod
    async def create_application(self) -> UvicornApplication:
        raise NotImplementedError

    async def start(self) -> Any:
        app = await self.create_application()

        config = Config(
            app,
            host=self.host,
            port=self.port,
            uds=self.uds,
            fd=self.fd,
            http=self.http,
            ws=self.ws,
            ws_max_size=self.ws_max_size,
            ws_ping_interval=self.ws_ping_interval,
            ws_ping_timeout=self.ws_ping_timeout,
            ws_per_message_deflate=self.ws_per_message_deflate,
            lifespan=self.lifespan,
            access_log=self.access_log,
            interface=self.interface,
            proxy_headers=self.proxy_headers,
            server_header=self.server_header,
            date_header=self.date_header,
            forwarded_allow_ips=self.forwarded_allow_ips,
            root_path=self.root_path,
            limit_concurrency=self.limit_concurrency,
            limit_max_requests=self.limit_max_requests,
            backlog=self.backlog,
            timeout_keep_alive=self.timeout_keep_alive,
            timeout_notify=self.timeout_notify,
            timeout_graceful_shutdown=self.timeout_graceful_shutdown,
            callback_notify=self.callback_notify,
            ssl_keyfile=self.ssl_keyfile,
            ssl_certfile=self.ssl_certfile,
            ssl_keyfile_password=self.ssl_keyfile_password,
            ssl_version=self.ssl_version,
            ssl_cert_reqs=self.ssl_cert_reqs,
            ssl_ca_certs=self.ssl_ca_certs,
            ssl_ciphers=self.ssl_ciphers,
            headers=self.headers,
            factory=self.factory,
            h11_max_incomplete_event_size=self.h11_max_incomplete_event_size,
            log_config=None,
            use_colors=False,
        )
        if not self.sock:
            self.sock = config.bind_socket()
        server = Server(config)

        self.start_event.set()

        await server.serve(sockets=[self.sock])



import asyncio
import atexit
import logging
import logging.handlers
import time
import traceback
from contextlib import suppress
from functools import partial
from socket import socket
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from weakref import finalize

import aiomisc_log
from aiomisc_log.enum import LogFormat, LogLevel

from .thread_pool import run_in_new_thread


def _thread_flusher(
    handler: logging.handlers.MemoryHandler,
    flush_interval: Union[float, int],
    loop: asyncio.AbstractEventLoop,
) -> None:
    def has_no_target() -> bool:
        return True

    def has_target() -> bool:
        return bool(handler.target)

    is_target = has_no_target

    if isinstance(handler, logging.handlers.MemoryHandler):
        is_target = has_target

    while not loop.is_closed() and is_target():
        with suppress(Exception):
            if handler.buffer:
                handler.flush()

        time.sleep(flush_interval)


def suppressor(
    callback: Callable[..., None],
    exceptions: Tuple[Type[BaseException], ...] = (Exception,),
) -> Callable[..., None]:
    def wrapper() -> None:
        with suppress(*exceptions):
            callback()
    return wrapper


def wrap_logging_handler(
    handler: logging.Handler,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    buffer_size: int = 1024,
    flush_interval: Union[float, int] = 0.1,
) -> logging.Handler:
    buffered_handler = logging.handlers.MemoryHandler(
        buffer_size,
        target=handler,
        flushLevel=logging.CRITICAL,
    )

    run_in_new_thread(
        _thread_flusher, args=(
            buffered_handler, flush_interval, loop,
        ), no_return=True, statistic_name="logger",
    )

    at_exit_flusher = suppressor(handler.flush)
    atexit.register(at_exit_flusher)
    finalize(buffered_handler, partial(atexit.unregister, at_exit_flusher))

    return buffered_handler


class UnhandledLoopHook(aiomisc_log.UnhandledHookBase):
    LOGGER_NAME = "asyncio.unhandled"

    @staticmethod
    def _fill_transport_extra(
        transport: Optional[asyncio.Transport],
        extra: Dict[str, Any],
    ) -> None:
        if transport is None:
            return

        extra["transport"] = repr(transport)

        for key in (
            "peername", "sockname", "compression",
            "cipher", "peercert", "pipe", "subprocess",
        ):
            value = transport.get_extra_info(key)
            if value:
                extra[f"transport_{key}"] = value

    def __call__(
        self, loop: asyncio.AbstractEventLoop,
        context: Dict[str, Any],
    ) -> None:
        context = dict(context)
        message: str = context.pop("message", "unhandled loop exception")
        exception: Optional[BaseException] = context.pop("exception", None)
        future: Optional[asyncio.Future] = context.pop("future", None)
        task: Optional[asyncio.Task] = context.pop("task", None)
        handle: Optional[asyncio.Handle] = context.pop("handle", None)
        protocol: Optional[asyncio.Protocol] = context.pop("protocol", None)
        transport: Optional[asyncio.Transport] = context.pop("transport", None)
        sock: Optional[socket] = context.pop("socket", None)
        source_traceback: List[traceback.FrameSummary] = context.pop(
            "source_traceback", None,
        )

        if exception is None:
            if future is not None:
                exception = future.exception()
            elif task is not None and task.done():
                exception = task.exception()

        extra = context
        if handle is not None:
            extra["handle"] = repr(handle)
        if protocol is not None:
            extra["protocol"] = repr(protocol)
        if sock is not None:
            extra["sock"] = repr(sock)

        self._fill_transport_extra(transport, extra)
        self.logger.exception(message, exc_info=exception, extra=extra)
        if source_traceback:
            self.logger.error(
                "".join(traceback.format_list(source_traceback)),
            )


def basic_config(
    level: Union[int, str] = LogLevel.default(),
    log_format: Union[str, LogFormat] = LogFormat.default(),
    buffered: bool = True, buffer_size: int = 1024,
    flush_interval: Union[int, float] = 0.2,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    **kwargs: Any,
) -> None:
    loop = loop or asyncio.get_event_loop()
    unhandled_hook = UnhandledLoopHook()

    def wrap_handler(handler: logging.Handler) -> logging.Handler:
        nonlocal buffer_size, buffered, loop, unhandled_hook

        unhandled_hook.set_handler(handler)

        if buffered:
            return wrap_logging_handler(
                handler=handler,
                buffer_size=buffer_size,
                flush_interval=flush_interval,
                loop=loop,
            )
        return handler

    aiomisc_log.basic_config(
        level=level,
        log_format=log_format,
        handler_wrapper=wrap_handler,
        **kwargs,
    )

    loop.set_exception_handler(unhandled_hook)


__all__ = (
    "LogFormat",
    "LogLevel",
    "basic_config",
)



import hashlib
import logging
import signal
import socket
import sys
from enum import IntEnum
from typing import Tuple, TypeVar, Union


T = TypeVar("T")
AddressType = Union[str, Tuple[str, int]]

log = logging.getLogger(__name__)

SALT_SIZE = 64
COOKIE_SIZE = 128
HASHER = hashlib.sha256


class PacketTypes(IntEnum):
    REQUEST = 0
    EXCEPTION = 1
    RESULT = 2
    CANCELLED = 3
    AUTH = 50
    AUTH_FAIL = 58
    AUTH_OK = 59
    BAD_INITIALIZER = 60
    BAD_PACKET = 254


INET_AF = socket.AF_INET6


if sys.platform in ("win32", "cygwin"):
    SIGNAL = signal.SIGBREAK      # type: ignore
    INT_SIGNAL = signal.SIGINT
else:
    SIGNAL = signal.SIGUSR2
    INT_SIGNAL = signal.SIGINT



import logging
import os
from itertools import chain
from types import MappingProxyType
from typing import Callable, Mapping

from aiomisc.compat import entry_pont_iterator


def setup_plugins() -> Mapping[str, Callable]:
    if os.getenv("AIOMISC_NO_PLUGINS"):
        return MappingProxyType({})

    plugins = {}
    logger = logging.getLogger(__name__)

    for entry_point in chain(
        entry_pont_iterator("aiomisc.plugins"),
        entry_pont_iterator("aiomisc"),
    ):
        try:
            plugins[entry_point.name] = entry_point.load()
        except:  # noqa
            logger.exception(
                "Failed to load entrypoint %r", entry_point,
            )

    for name, plugin in plugins.items():
        try:
            logger.debug("Trying to load %r %r", name, plugin)
            plugin.setup()
        except:  # noqa
            logger.exception("Error on %s aiomisc plugin setup", name)
            raise

    return MappingProxyType(plugins)


plugins: Mapping[str, Callable] = setup_plugins()


__all__ = ("plugins",)



import cProfile
import io
import logging
from pstats import Stats
from typing import Optional

from ..periodic import PeriodicCallback
from .base import Service


class Profiler(Service):
    profiler = None        # type: cProfile.Profile
    periodic = None        # type: PeriodicCallback

    order = "cumulative"    # type: str

    path = None             # type: str
    logger = None           # type: logging.Logger

    interval = 10           # type: int
    top_results = 10        # type: int
    log = logging.getLogger(__name__)   # type: logging.Logger

    async def start(self) -> None:
        self.logger = self.log.getChild(str(id(self)))

        self.profiler = cProfile.Profile()
        self.periodic = PeriodicCallback(self.save_stats)

        self.profiler.enable()
        self.periodic.start(self.interval)

    def save_stats(self) -> None:
        with io.StringIO() as stream:
            stats = Stats(
                self.profiler, stream=stream,
            ).strip_dirs().sort_stats(self.order)

            stats.print_stats(self.top_results)
            self.logger.info(stream.getvalue())

            try:
                if self.path is not None:
                    stats.dump_stats(self.path)
            finally:
                self.profiler.enable()

    async def stop(self, exception: Optional[Exception] = None) -> None:
        self.logger.info("Stop profiler")
        await self.periodic.stop(return_exceptions=True)
        self.profiler.disable()



import asyncio
import io
import pickle
import platform
import socket
from abc import ABC, abstractmethod
from struct import Struct
from typing import IO, Any, BinaryIO, Union


ADDRESS_FAMILY = (
    socket.AF_INET6 if platform.system() == "Windows" else socket.AF_UNIX
)


class Protocol(ABC):
    PACKET_HEADER = Struct("!L")

    @abstractmethod
    def _read(self, size: int) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def _write(self, data: bytes) -> None:
        raise NotImplementedError

    def _pack(self, payload: Any) -> bytes:
        data = pickle.dumps(payload)
        return self.PACKET_HEADER.pack(len(data)) + data

    def send(self, payload: Any) -> None:
        self._write(self._pack(payload))

    def receive(self) -> Any:
        header = self._read(self.PACKET_HEADER.size)
        payload_size = self.PACKET_HEADER.unpack(header)[0]
        payload_bytes = self._read(payload_size)
        return pickle.loads(payload_bytes)


class FileIOProtocol(Protocol):
    fd: Union[BinaryIO, IO[bytes]]

    def __init__(self, fd: Union[BinaryIO, IO[bytes]]):
        self.fd = fd

    def _read(self, size: int) -> bytes:
        return self.fd.read(size)

    def _write(self, data: bytes) -> None:
        self.fd.write(data)


class SocketIOProtocol(Protocol):
    sock: socket.socket

    def __init__(self, sock: socket.socket):
        self.sock = sock

    def _read(self, size: int) -> bytes:
        with io.BytesIO() as fp:
            while fp.tell() < size:
                chunk = self.sock.recv(size - fp.tell())
                if len(chunk) == 0:
                    raise ConnectionError("Remote didn't send any data")
                fp.write(chunk)
            return fp.getvalue()

    def _write(self, data: bytes) -> None:
        self.sock.send(data)


class AsyncProtocol:
    PACKET_HEADER = Protocol.PACKET_HEADER

    def __init__(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter,
    ):
        self.reader = reader
        self.writer = writer

    def _pack(self, payload: Any) -> bytes:
        data = pickle.dumps(payload)
        return self.PACKET_HEADER.pack(len(data)) + data

    async def send(self, payload: Any) -> None:
        self.writer.write(self._pack(payload))
        await self.writer.drain()

    async def receive(self) -> Any:
        header = await self.reader.readexactly(self.PACKET_HEADER.size)
        payload_size = self.PACKET_HEADER.unpack(header)[0]
        payload_bytes = await self.reader.readexactly(payload_size)
        return pickle.loads(payload_bytes)

    def close(self) -> None:
        if not self.writer.can_write_eof():
            self.writer.write_eof()



import inspect
from typing import Any, Callable, FrozenSet, Set, TypeVar, Union


T = TypeVar("T")
ReceiverType = Callable[..., Any]
_ReceiverSetType = Union[Set[ReceiverType], FrozenSet[ReceiverType]]


class Signal:

    __slots__ = ("_receivers", "_is_frozen")

    def __init__(self) -> None:
        self._receivers = set()   # type: _ReceiverSetType

    def connect(self, receiver: ReceiverType) -> None:
        if self.is_frozen:
            raise RuntimeError(
                "Can't connect receiver (%r) to the frozen signal",
                receiver,
            )

        if not inspect.iscoroutinefunction(receiver):
            raise RuntimeError("%r is not a coroutine function", receiver)

        self._receivers.add(receiver)   # type: ignore

    def disconnect(self, receiver: ReceiverType) -> None:
        if self.is_frozen:
            raise RuntimeError(
                "Can't connect receiver (%r) to the frozen signal",
                receiver,
            )

        self._receivers.discard(receiver)  # type: ignore

    async def call(self, *args: Any, **kwargs: Any) -> None:
        for receiver in self._receivers:
            await receiver(*args, **kwargs)

    def copy(self) -> "Signal":
        clone = Signal()
        # unfreeze on copy
        clone._receivers = set(self._receivers)
        return clone

    @property
    def is_frozen(self) -> bool:
        return isinstance(self._receivers, frozenset)

    def freeze(self) -> None:
        self._receivers = frozenset(self._receivers)


def receiver(s: Signal) -> Callable[..., Callable[..., T]]:
    def decorator(func: Callable[..., T]) -> Callable[..., T]:
        s.connect(func)
        return func

    return decorator



import asyncio
import logging
from typing import Any, Optional, Union

from .compat import EventLoopMixin
from .recurring import CallbackType, ExceptionsType, RecurringCallback


log = logging.getLogger(__name__)


class PeriodicCallback(EventLoopMixin):
    """
    .. note::

        When the periodic function executes longer then execution interval a
        next call would be skipped and warning would be logged.

    """

    __slots__ = ("_recurring_callback", "_task") + EventLoopMixin.__slots__

    _task: Optional[asyncio.Future]

    def __init__(
        self, coroutine_func: CallbackType, *args: Any, **kwargs: Any,
    ):
        self._recurring_callback: RecurringCallback = RecurringCallback(
            coroutine_func, *args, **kwargs,
        )
        self._task: Optional[asyncio.Task] = None

    def start(
        self, interval: Union[int, float],
        loop: Optional[asyncio.AbstractEventLoop] = None, *,
        delay: Union[int, float] = 0,
        shield: bool = False,
        suppress_exceptions: ExceptionsType = (),
    ) -> None:
        assert interval

        if self._task and not self._task.done():
            raise asyncio.InvalidStateError

        delayed = False

        def strategy(_: Any) -> Union[int, float]:
            nonlocal delayed
            if not delayed:
                delayed = True
                return delay
            return interval

        self._task = self._recurring_callback.start(
            strategy=strategy,
            shield=shield,
            loop=loop,
            suppress_exceptions=suppress_exceptions,
        )

        def clean_task(_: Any) -> None:
            self._task = None

        self._task.add_done_callback(clean_task)

    def stop(self, return_exceptions: bool = False) -> asyncio.Future:
        if self._task is None:
            self._task = self.loop.create_future()
            self._task.set_exception(RuntimeError("Callback not started"))
        elif not self._task.done():
            self._task.cancel()
        task, self._task = self._task, None
        return asyncio.gather(task, return_exceptions=return_exceptions)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._recurring_callback.name})"
